#!/usr/bin/env python
# coding=utf-8
#-----------------------------------------------------------------------
# Copyright © 2014-2024 Tormach® Inc. All rights reserved.
# License: GPL Version 2
#-----------------------------------------------------------------------

# if this file exists in ~/gcode/python/ui_hooks.py the UI will import it
import time
import threading
import traceback
import os
import urllib3
import linuxcnc
import redis
import constants
from pp_hub.connector import Connector, TokenError, TransferItem

# not great to do this, but PathPilot's OS is old enough that connecting to Hub 
# generates SSL warnings in th log.  SSL is still used, it just complains it has 
# to downgrade to a "less-secure" SSL and suggests upgrading Python, which isn't
# exactly an option.
urllib3.disable_warnings()

class ui_hooks():

    TIMER_FAST = 2
    TIMER_SLOW = 30
    GCODE_PATH = os.path.join(os.environ['HOME'], 'gcode')

    def truncate(self, f, n):
        '''Truncates/pads a float f to n decimal places without rounding'''
        s = '{}'.format(f)
        if 'e' in s or 'E' in s:
            return '{0:.{1}f}'.format(f, n)
        i, p, d = s.partition('.')
        return '.'.join([i, (d+'0'*n)[:n]])

    def flatten_cloud_files(self, dirname, file_list):
        '''
        Recursivley flatten the response from PathPilot's Connector to Hub into
        a similar <filepath>:<modified> structure to get_local_files()
        '''
        result = []
        for item in file_list:
            item_type = item.get('type')
            if item_type == 'f':
                result.append((
                    os.path.join(dirname, item['name']), 
                    self.connector.timestamp_for_datetime(
                        self.connector.cloudutc_to_local_datetime(item['mtime_utc'])
                    )
                ))
            elif item_type == 'd':
                result.extend(self.flatten_cloud_files(os.path.join(dirname, item['name']), item['children']))
        return result

    def get_cloud_files(self):
        '''Return a dict of <filepath>:<modified> pairs from your hub accont'''
        cloud_files = self.connector.get_cloud_files()
        file_list = self.flatten_cloud_files('', cloud_files['contents'])
        result = {}
        for file in file_list:
            result[file[0]] = float(self.truncate(file[1], 3))
        return result

    def get_local_files(self):
        '''Return a dict of <filepath>:<modified> pairs from local gcode folder'''
        result = {}
        for dirpath, _, files in os.walk(self.GCODE_PATH):
            for file in files:
                filepath = os.path.join(dirpath, file)
                stat = os.stat(filepath)
                result[filepath.replace(self.GCODE_PATH, 'gcode')] = float(
                    self.truncate(stat.st_mtime, 3)
                )
        return result

    def hubsync(self):
        while True:
            try:
                # Get linuxcnc status, and get logged in state
                self.stat.poll()
                token = self.redis.hget('pp_hub_settings', 'token')

                # If running a program, do nothing and sleep for TIMER_SLOW sec
                if self.stat.task_state == linuxcnc.STATE_ON and self.stat.task_mode == linuxcnc.MODE_AUTO:
                    self.error_handler.write(
                        'ui_hooks:hubsync() -- gcode program running, long pause',
                        constants.ALARM_LEVEL_DEBUG
                    )
                    time.sleep(self.TIMER_SLOW)
                elif not token:
                    # not signed in, in case the user has signed out, invalidate connector
                    # and wait TIMER_SLOW to check again if they have signed in
                    self.error_handler.write(
                        'ui_hooks:hubsync() -- user not signed in, long pause',
                        constants.ALARM_LEVEL_DEBUG
                    )
                    if self.connector: self.connector = None
                    time.sleep(self.TIMER_SLOW)
                else:
                    # signed in, ensure we have a connector related to sign in session
                    if not self.connector:
                        self.connector = Connector(constants.PP_HUB_URL, self.redis)
                    # Get Local filelist
                    local_files = self.get_local_files()
                    # Get Hub filelist
                    cloud_files = self.get_cloud_files()

                    # Download any newer file on Hub
                    download_list = []
                    for file in cloud_files:
                        if file not in local_files:
                            self.error_handler.write(
                                'ui_hooks:hubsync() -- File on hub but not local:' + file,
                                constants.ALARM_LEVEL_DEBUG
                            )
                            download_list.append(TransferItem(
                                os.path.join(os.environ['HOME'], file),
                                file,
                                'f',
                                None,
                                None
                            ))
                        elif local_files[file] < cloud_files[file]:
                            self.error_handler.write(
                                'ui_hooks:hubsync() -- Newer file on hub: %s (%.3f ltime, %.3f ctime)' % (
                                    file, local_files[file], cloud_files[file]
                                ), constants.ALARM_LEVEL_DEBUG
                            )
                            download_list.append(TransferItem(
                                os.path.join(os.environ['HOME'], file),
                                file,
                                'f',
                                None,
                                None
                            ))
                    if download_list:
                        self.error_handler.write('ui_hooks:hubsync() -- downloading', constants.ALARM_LEVEL_DEBUG)
                        self.connector.download_cloud_items(download_list)

                    # Upload any file newer than on hub
                    upload_list = []
                    for file in local_files:
                        if file in cloud_files:
                            if local_files[file] > cloud_files[file]:
                                self.error_handler.write(
                                    'ui_hooks:hubsync() -- Newer file local: %s (%.3f ltime, %.3f ctime)' % (
                                        file, local_files[file], cloud_files[file]
                                    ), constants.ALARM_LEVEL_DEBUG
                                )
                                upload_list.append(TransferItem(
                                    os.path.join(os.environ['HOME'], file),
                                    file,
                                    'f',
                                    None,
                                    None
                                ))
                    if upload_list:
                        self.error_handler.write('ui_hooks:hubsync() -- uploading', constants.ALARM_LEVEL_DEBUG)
                        self.connector.upload_cloud_items(upload_list, False)
                    # check again in TIMER_FAST seconds
                    time.sleep(self.TIMER_FAST)
            except TokenError:
                # Despite having a token, it was invalid, invalidate connector
                # and wait TIMER_SLOW to check again if they have signed in
                self.error_handler.write(
                    'ui_hooks:hubsync() -- user not signed in, long pause',
                    constants.ALARM_LEVEL_DEBUG
                )
                if self.connector: self.connector = None
                time.sleep(self.TIMER_SLOW)
            except Exception:
                # Some other malfunction, but just in case it's temporary (lost internet?)
                # wait TIMER_SLOW before retrying
                self.error_handler.write('ui_hooks:hubsync() -- Exception', constants.ALARM_LEVEL_HIGH)
                self.error_handler.write(traceback.format_exc(), constants.ALARM_LEVEL_HIGH)
                time.sleep(self.TIMER_SLOW)


    def __init__(self, linuxcnc_command, error_handler, ver_list, digital_output_offset=0):
        # log that we were found
        try:
            self.command = linuxcnc_command
            self.stat = linuxcnc.stat()
            self.error_handler = error_handler
            self.error_handler.write('ui_hooks:__init__() -- setup hubsync', constants.ALARM_LEVEL_DEBUG)
            self.redis = redis.Redis()
            self.connector = None
            self.signed_in = False
            self.thread = threading.Thread(target=self.hubsync)
            self.thread.daemon = True
            self.thread.start()
        except Exception as ex:
            self.error_handler.write('ui_hooks:__init__() -- failed hubsync setup', constants.ALARM_LEVEL_HIGH)
            self.error_handler.write(str(ex), constants.ALARM_LEVEL_HIGH)
